import unittest

from src.otyg_risk_base.quantitative_risk import QuantitativeRisk


class TestQuantitativeRisk(unittest.TestCase):
    def test_no_arg(self):
        risk = QuantitativeRisk()
        self.assertIsInstance(risk, QuantitativeRisk)
        for s in [
            "threat_event_frequency",
            "vulnerability",
            "loss_event_frequency",
            "loss_magnitude",
            "annual_loss_expectancy",
            "budget",
            "currency",
        ]:
            self.assertIn(s, risk.to_dict())

    def test_equality(self):
        risk = QuantitativeRisk()
        self.assertTrue(risk == risk)

    def test_serialization_deserialization(self):
        risk = QuantitativeRisk()
        new_risk = QuantitativeRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == new_risk)
        risk = QuantitativeRisk(
            {
                "threat_event_frequency": {"min": 0, "probable": 1, "max": 2},
                "vulnerability": {"min": 1, "probable": 2, "max": 3},
                "loss_magnitude": {"min": 3, "probable": 4, "max": 5},
                "budget": 10000,
                "currency": "SEK",
            }
        )
        new_risk_2 = QuantitativeRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == new_risk_2)
        self.assertFalse(risk == new_risk)


if __name__ == "__main__":
    unittest.main()



from .qualitative_scale import QualitativeScale
from .utils import *


class QualitativeRisk:
    def __init__(
        self,
        likelihood_init: int = 1,
        likelihood_impact: int = 1,
        impact: int = 1,
        mappings: QualitativeScale = None,
    ):
        if not mappings:
            self.mappings = QualitativeScale()
        else:
            self.mappings = mappings
        raw_likelihood = self.mappings.get(
            raw=likelihood_impact * likelihood_init, mapping="risk"
        )
        self.overall_likelihood: str = raw_likelihood.get("text")
        self.overall_likelihood_num: int = raw_likelihood.get("numeric")
        self.likelihood_initiation_or_occurence: str = self.get_text_from_num(
            likelihood_init
        )
        self.likelihood_adverse_impact: str = self.get_text_from_num(likelihood_impact)
        self.impact: str = self.get_text_from_num(impact)
        self.overall_risk: str = self.mappings.get(
            raw=self.overall_likelihood_num * impact, mapping="risk"
        ).get("text")

    def get_text_from_num(self, num):
        value = ""
        try:
            value = self.mappings.num_to_text[num]
        except KeyError:
            try:
                value = self.mappings.num_to_text[str(num)]
            except:
                raise
        return value

    def get(self):
        return {
            "risk": self.overall_risk,
            "likelihood": self.overall_likelihood,
            "impact": self.impact,
        }

    def to_dict(self):
        risk = {
            "overall_likelihood": self.overall_likelihood,
            "overall_likelihood_num": self.overall_likelihood_num,
            "likelihood_initiation_or_occurence": self.likelihood_initiation_or_occurence,
            "likelihood_adverse_impact": self.likelihood_adverse_impact,
            "impact": self.impact,
            "overall_risk": self.overall_risk,
            "mappings": self.mappings.to_dict(),
        }
        return risk

    @classmethod
    def from_dict(cls, values: dict):
        risk = QualitativeRisk()
        risk.overall_likelihood = values.get("overall_likelihood")
        risk.overall_likelihood_num = values.get("overall_likelihood_num")
        risk.likelihood_initiation_or_occurence = values.get(
            "likelihood_initiation_or_occurence"
        )
        risk.likelihood_adverse_impact = values.get("likelihood_adverse_impact")
        risk.impact = values.get("impact")
        risk.overall_risk = values.get("overall_risk")
        risk.mappings = QualitativeScale.from_dict(scales=values.get("mappings"))
        return risk

    def __hash__(self):
        return hash(freeze(self.to_dict()))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()



from decimal import Decimal
import unittest

from src.otyg_risk_base.montecarlo import MonteCarloRange
from src.otyg_risk_base.hybrid import HybridRisk


class TestHybridRisk(unittest.TestCase):
    def test_set_qualitative(self):
        risk = HybridRisk(
            {
                "threat_event_frequency": {"min": 0, "probable": 1, "max": 2},
                "vulnerability": {"min": 0, "probable": 2, "max": 3},
                "loss_magnitude": {"min": 1, "probable": 2, "max": 3},
                "budget": 10000,
                "currency": "SEK",
            }
        )
        self.assertEqual(risk.qualitative.overall_risk, "Very High")
        self.assertEqual(risk.qualitative.overall_likelihood, "High")
        self.assertEqual(risk.qualitative.impact, "Very High")

    def test_equality(self):
        risk = HybridRisk(
            {
                "threat_event_frequency": {"min": 0, "probable": 1, "max": 2},
                "vulnerability": {"min": 1, "probable": 2, "max": 3},
                "loss_magnitude": {"min": 3, "probable": 4, "max": 5},
                "budget": 10000,
                "currency": "SEK",
            }
        )
        self.assertTrue(risk == risk)
        risk_mod = HybridRisk(
            {
                "threat_event_frequency": {"min": 1, "probable": 3, "max": 5},
                "vulnerability": {"min": 1, "probable": 2, "max": 3},
                "loss_magnitude": {"min": 3, "probable": 4, "max": 5},
                "budget": 10000,
                "currency": "SEK",
            }
        )
        self.assertFalse(risk == risk_mod)

    def test_serialization_deserialization(self):
        risk = HybridRisk(
            {
                "threat_event_frequency": MonteCarloRange(min=0, probable=1, max=2),
                "vulnerability": MonteCarloRange(min=1, probable=2, max=3),
                "loss_magnitude": MonteCarloRange(min=4, probable=5, max=6),
                "budget": 10000,
                "currency": "SEK",
            }
        )
        risk_new = HybridRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == risk_new)
        risk = HybridRisk(
            {
                "threat_event_frequency": {"min": 0, "probable": 1, "max": 2},
                "vulnerability": {"min": 1, "probable": 2, "max": 3},
                "loss_magnitude": {"min": 3, "probable": 4, "max": 5},
                "budget": 10000,
                "currency": "SEK",
            }
        )
        risk_new = HybridRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == risk_new)
        risk_mod = HybridRisk(
            {
                "threat_event_frequency": {"min": 1, "probable": 3, "max": 5},
                "vulnerability": {"min": 1, "probable": 2, "max": 3},
                "loss_magnitude": {"min": 3, "probable": 4, "max": 5},
                "budget": 10000,
                "currency": "SEK",
            }
        )
        risk_new = HybridRisk.from_dict(risk_mod.to_dict())
        self.assertFalse(risk == risk_new)
        self.assertTrue(risk_mod == risk_new)

    def test_ui_bugg(self):
        values = {
            "threat_event_frequency": {
                "min": 0.023,
                "probable": 0.12250000000000001,
                "max": 0.222,
            },
            "vulnerability": {"min": 0.0, "probable": 0.085, "max": 0.17},
            "loss_magnitude": {"min": 0.005, "probable": 0.01, "max": 0.02},
            "budget": Decimal("1000000"),
            "currency": "SEK",
            "mappings": {
                "likelihood_initiation_or_occurence": [
                    {"value": 1, "low": 0.01, "high": 0.1},
                    {"value": 2, "low": 0.1, "high": 1},
                    {"value": 3, "low": 1, "high": 10},
                    {"value": 4, "low": 10, "high": 100},
                    {"value": 5, "low": 100, "high": 1000},
                ],
                "likelihood_adverse_impact": [
                    {"value": 1, "low": 0.01, "high": 0.12},
                    {"value": 2, "low": 0.12, "high": 0.25},
                    {"value": 3, "low": 0.25, "high": 0.5},
                    {"value": 4, "low": 0.5, "high": 0.75},
                    {"value": 5, "low": 0.75, "high": 1},
                ],
                "impact": [
                    {"value": 1, "low": 0.0001, "high": 0.005},
                    {"value": 2, "low": 0.005, "high": 0.01},
                    {"value": 3, "low": 0.01, "high": 0.02},
                    {"value": 4, "low": 0.02, "high": 0.05},
                    {"value": 5, "low": 0.05, "high": 1},
                ],
                "num_to_text": {
                    "5": "Very High",
                    "4": "High",
                    "3": "Moderate",
                    "2": "Low",
                    "1": "Very Low",
                    "0": "Very Low",
                },
                "risk": [
                    {"value": 1, "low": 1, "high": 5},
                    {"value": 2, "low": 5, "high": 9},
                    {"value": 3, "low": 9, "high": 13},
                    {"value": 4, "low": 13, "high": 20},
                    {"value": 5, "low": 20, "high": 26},
                ],
            },
        }
        risk = HybridRisk(values=values)
        self.assertIsNotNone(risk.qualitative.overall_risk)
        values = {
            "threat_event_frequency": {"min": 0.0, "probable": 0.0, "max": 0.0},
            "vulnerability": {"min": 0.0, "probable": 0.0, "max": 0.0},
            "loss_magnitude": {"min": 0.0, "probable": 0.0, "max": 0.0},
            "budget": Decimal("1000000"),
            "currency": "SEK",
            "mappings": {
                "likelihood_initiation_or_occurence": [
                    {"value": 1, "low": 0.01, "high": 0.1},
                    {"value": 2, "low": 0.1, "high": 1},
                    {"value": 3, "low": 1, "high": 10},
                    {"value": 4, "low": 10, "high": 100},
                    {"value": 5, "low": 100, "high": 1000},
                ],
                "likelihood_adverse_impact": [
                    {"value": 1, "low": 0.01, "high": 0.12},
                    {"value": 2, "low": 0.12, "high": 0.25},
                    {"value": 3, "low": 0.25, "high": 0.5},
                    {"value": 4, "low": 0.5, "high": 0.75},
                    {"value": 5, "low": 0.75, "high": 1},
                ],
                "impact": [
                    {"value": 1, "low": 0.0001, "high": 0.005},
                    {"value": 2, "low": 0.005, "high": 0.01},
                    {"value": 3, "low": 0.01, "high": 0.02},
                    {"value": 4, "low": 0.02, "high": 0.05},
                    {"value": 5, "low": 0.05, "high": 1},
                ],
                "num_to_text": {
                    "5": "Very High",
                    "4": "High",
                    "3": "Moderate",
                    "2": "Low",
                    "1": "Very Low",
                    "0": "Very Low",
                },
                "risk": [
                    {"value": 1, "low": 1, "high": 5},
                    {"value": 2, "low": 5, "high": 9},
                    {"value": 3, "low": 9, "high": 13},
                    {"value": 4, "low": 13, "high": 20},
                    {"value": 5, "low": 20, "high": 26},
                ],
            },
        }
        risk = HybridRisk(values=values)
        self.assertIsNotNone(risk.qualitative.overall_risk)



import unittest

from src.otyg_risk_base.qualitative_scale import QualitativeScale


class TestQualitativeScale(unittest.TestCase):
    def test_get(self):
        scale = QualitativeScale()
        self.assertEqual(
            scale.get(raw=0.0001, mapping="likelihood_initiation_or_occurence").get(
                "numeric"
            ),
            1,
        )
        self.assertEqual(
            scale.get(raw=0.1, mapping="likelihood_initiation_or_occurence").get(
                "numeric"
            ),
            2,
        )
        self.assertEqual(
            scale.get(raw=0.1, mapping="likelihood_initiation_or_occurence").get(
                "text"
            ),
            "Low",
        )
        self.assertEqual(scale.get(raw=0.06, mapping="impact").get("text"), "Very High")
        self.assertEqual(
            scale.get(raw=0.74, mapping="likelihood_adverse_impact").get("text"), "High"
        )
        self.assertEqual(scale.get(raw=16, mapping="risk").get("text"), "High")


if __name__ == "__main__":
    unittest.main()



#
# BSD 3-Clause License
#
# Copyright (c) 2026, Martin Vesterlund
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import numpy as np
from decimal import Decimal
import numpy as np
import math


class MonteCarloRange:
    def __init__(
        self,
        min: Decimal = Decimal(0.00),
        probable: Decimal = Decimal(0.00),
        max: Decimal = Decimal(0.00),
    ):
        if not (max == min) and ((probable < min) or (probable > max) or (min > max)):
            raise ValueError
        self.max = Decimal(max)
        self.min = Decimal(min)
        self.probable = Decimal(probable)
        if not self.probable.is_zero() and (self.max == self.min):
            self.max = Decimal(self.probable * 2)
            self.min = Decimal(self.probable / 2)
        elif self.probable.is_zero and not self.max.is_zero():
            self.probable = (self.min + self.max) / 2

    def to_dict(self):
        return {
            "min": float(self.min),
            "probable": float(self.probable),
            "max": float(self.max),
        }

    def add(self, other=None):
        if isinstance(other, MonteCarloRange):
            max = self.max + other.max
            min = self.min + other.min
            probable = self.probable + other.probable
        else:
            max = self.max + other
            min = self.min + other
            probable = self.probable + other
        result = MonteCarloRange(
            min=Decimal(min), probable=Decimal(probable), max=Decimal(max)
        )
        return result

    def sub(self, other=None):
        if isinstance(other, MonteCarloRange):
            max = self.max - other.max
            min = self.min - other.min
            probable = self.probable - other.probable
        else:
            max = self.max - other
            min = self.min - other
            probable = self.probable - other
        if max < min:
            tmp = max
            max = min
            min = tmp
        result = MonteCarloRange(
            min=Decimal(min), probable=Decimal(probable), max=Decimal(max)
        )
        return result

    def multiply(self, other=None):
        if isinstance(other, MonteCarloRange) or isinstance(
            other, MonteCarloSimulation
        ):
            max = self.max * other.max
            min = self.min * other.min
            probable = self.probable * other.probable
        else:
            max = Decimal(self.max) * Decimal(other)
            min = Decimal(self.min) * Decimal(other)
            probable = Decimal(self.probable) * Decimal(other)
        if min < 1e-20:
            min = Decimal(0.0)
        if probable < 1e-20:
            probable = Decimal(0.0)
        if max < 1e-20:
            max = Decimal(0.0)
        result = MonteCarloRange(
            min=Decimal(min), probable=Decimal(probable), max=Decimal(max)
        )
        return result

    @classmethod
    def from_dict(cls, values: dict = None):
        if isinstance(values, dict):
            return MonteCarloRange(
                min=values["min"], probable=values["probable"], max=values["max"]
            )
        elif isinstance(values, MonteCarloRange):
            return values
        else:
            raise TypeError

    def __repr__(self):
        return str(self.to_dict())

    def __hash__(self):
        return hash((self.min, self.max, self.probable))

    def __eq__(self, other):
        if isinstance(other, MonteCarloRange) and self.__hash__() == other.__hash__():
            return True
        return False

    def __gt__(self, other):
        if self == other:
            return False
        if self.min >= other.min and self.max > other.max:
            return True
        return False

class PertDistribution:
    def __init__(self, range: MonteCarloRange, samples: int = 100000):
        rng = np.random.default_rng()
        delta_min_max = range.max - range.min
        alpha = 1 + ((range.probable - range.min) * 4) / delta_min_max
        beta = 1 + ((range.max - range.probable) * 4) / delta_min_max
        self.__samples = float(range.min) + rng.beta(alpha, beta, samples) * float(
            delta_min_max
        )

    def get(self):
        return self.__samples.copy()


class LogLogisticDistribution:
    def __init__(
        self,
        range: MonteCarloRange,
        samples: int = 100000,
        upper_quantile: float = 0.99,
    ):
        rng = np.random.default_rng()

        min_value = float(range.min)
        probable_value = float(range.probable)
        max_value = float(range.max)

        eps = np.finfo(float).eps

        if max_value <= min_value:
            self.__samples = np.full(samples, min_value, dtype=float)
            return

        # Shifted log-logistic:
        # X = min_value + Y
        # Y ~ LogLogistic(alpha, beta)
        #
        # Median(X) = min_value + alpha
        alpha = probable_value - min_value
        if alpha <= eps:
            alpha = max((max_value - min_value) * 0.1, eps)

        span_ratio = (max_value - min_value) / alpha

        if span_ratio <= 1.0 + eps:
            beta = 10.0
        else:
            # Q(u) = min_value + alpha * (u / (1-u)) ** (1/beta)
            # Solve Q(upper_quantile) = max_value for beta
            logit_q = math.log(upper_quantile / (1.0 - upper_quantile))
            beta = logit_q / math.log(span_ratio)
            if not np.isfinite(beta) or beta <= 0:
                beta = 10.0

        # Truncate distribution to [min_value, max_value]
        # CDF(max_value) for shifted log-logistic
        f_max = 1.0 / (1.0 + (alpha / (max_value - min_value)) ** beta)

        # Sample uniformly in truncated CDF interval
        u = rng.uniform(low=eps, high=max(f_max - eps, eps), size=samples)

        # Inverse CDF
        self.__samples = min_value + alpha * (u / (1.0 - u)) ** (1.0 / beta)

    def get(self):
        return self.__samples.copy()

class MonteCarloSimulation:
    def __init__(self, range: MonteCarloRange = None):
        if not range:
            range = MonteCarloRange(probable=Decimal(1))
        if range.min == range.max == range.probable:
            self.probable = Decimal(range.probable)
            range.max = Decimal(Decimal(range.probable) + Decimal(0.000000000001))
            range.min = Decimal(Decimal(range.probable) - Decimal(0.000000000001))

        pd = LogLogisticDistribution(range=range)
        self.__samples = pd.get()
        self.probable = Decimal(np.mean(self.__samples))
        self.p90 = Decimal(np.percentile(self.__samples, 90))
        self.p75 = Decimal(np.percentile(self.__samples, 75))
        self.max = Decimal(np.max(self.__samples))
        self.min = Decimal(np.min(self.__samples))

    def to_dict(self):
        return {
            "min": float(self.min),
            "probable": float(self.probable),
            "max": float(self.max),
            "p90": float(self.p90),
            "p75": float(self.p75),
            "__samples": self.__samples,
        }

    @classmethod
    def from_dict(cls, values: dict = {}):
        if isinstance(values, dict) and "p90" in values:
            mcs = MonteCarloSimulation()
            mcs.min = Decimal(values["min"])
            mcs.probable = Decimal(values["probable"])
            mcs.max = Decimal(values["max"])
            mcs.p90 = Decimal(values["p90"])
            mcs.p75 = Decimal(values["p75"])
            mcs.__samples = np.array(values.get("__samples", []))
            return mcs
        else:
            range = MonteCarloRange.from_dict(values)
            return MonteCarloSimulation(range=range)

    def get_montecarlo_range(self):
        return MonteCarloRange(min=self.min, probable=self.probable, max=self.max)

    def multiply(self, other):
        if isinstance(other, MonteCarloSimulation):
            range = self.get_montecarlo_range().multiply(
                other=other.get_montecarlo_range()
            )
        else:
            range = self.get_montecarlo_range().multiply(other=other)
        return MonteCarloSimulation(range=range)

    def add(self, other):
        if isinstance(other, MonteCarloSimulation):
            range = self.get_montecarlo_range().add(other=other.get_montecarlo_range())
        else:
            range = self.get_montecarlo_range().add(other=other)
        return MonteCarloSimulation(range=range)

    def delta(self, other):
        if isinstance(other, MonteCarloSimulation):
            range = self.get_montecarlo_range().sub(other=other.get_montecarlo_range())
        else:
            range = self.get_montecarlo_range().sub(other=other)
        return MonteCarloSimulation(range=range)

    def __add__(self, other):
        return self.add(other)

    def __radd__(self, other):
        return self.add(other)

    def __mul__(self, other):
        return self.multiply(other)

    def __rmul__(self, other):
        return self.multiply(other)

    def __repr__(self):
        return str(self.to_dict())

    def __str__(self):
        return f"min: {str(self.min)} mode: {str(self.probable)} p75: {str(self.p75)} p90: {str(self.p90)} max: {str(self.max)}"

    def __hash__(self):
        return hash((self.min, self.probable, self.p75, self.p90, self.max))

    def __eq__(self, value):
        if (
            isinstance(value, MonteCarloSimulation)
            and self.__hash__() == value.__hash__()
        ):
            return True
        return False

    def __gt__(self, other):
        if self == other:
            return False
        if self.min >= other.min and self.max > other.max:
            return True
        return False



#
# BSD 3-Clause License
#
# Copyright (c) 2026, Martin Vesterlund
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from decimal import *

from .montecarlo import MonteCarloRange, MonteCarloSimulation


class QuantitativeRisk:
    def __init__(self, values: dict = None):
        if not values:
            self.threat_event_frequency = MonteCarloRange()
            self.vuln_score = MonteCarloRange()
            self.loss_magnitude = MonteCarloSimulation()
            self.budget = Decimal(1000000)
            self.currency = "SEK"
        else:
            self.threat_event_frequency = MonteCarloRange.from_dict(
                values["threat_event_frequency"]
            )
            self.vuln_score = MonteCarloRange.from_dict(values["vulnerability"])
            self.loss_magnitude = MonteCarloSimulation.from_dict(
                values.get("loss_magnitude")
            )
            self.budget = Decimal(values["budget"])
            self.currency = values["currency"]
        if values and "loss_event_frequency" in values:
            self.loss_event_frequency = MonteCarloSimulation.from_dict(
                values.get("loss_event_frequency")
            )
            if "ale" in values and "annual_loss_expectancy" in values:
                self.ale = MonteCarloSimulation.from_dict(values.get("ale"))
                self.annual_loss_expectancy = MonteCarloSimulation.from_dict(
                    values.get("annual_loss_expectancy")
                )
            else:
                self.update_ale()
        else:
            self.loss_event_frequency = MonteCarloSimulation(
                self.threat_event_frequency.multiply(self.vuln_score)
            )
            self.update_ale()

    def update_ale(self):
        self.ale = self.loss_event_frequency.multiply(self.loss_magnitude).multiply(
            self.budget
        )
        self.annual_loss_expectancy = MonteCarloSimulation(self.ale)

    def to_dict(self):
        return {
            "threat_event_frequency": self.threat_event_frequency.to_dict(),
            "vulnerability": self.vuln_score.to_dict(),
            "loss_event_frequency": self.loss_event_frequency.to_dict(),
            "loss_magnitude": self.loss_magnitude.to_dict(),
            "ale": self.ale.to_dict(),
            "annual_loss_expectancy": self.annual_loss_expectancy.to_dict(),
            "budget": float(self.budget),
            "currency": self.currency,
        }

    @classmethod
    def from_dict(cls, values):
        risk = QuantitativeRisk()
        risk.threat_event_frequency = MonteCarloRange.from_dict(
            values["threat_event_frequency"]
        )
        risk.vuln_score = MonteCarloRange.from_dict(values["vulnerability"])
        risk.loss_magnitude = MonteCarloSimulation.from_dict(
            values.get("loss_magnitude")
        )
        risk.budget = Decimal(values["budget"])
        risk.currency = values["currency"]
        risk.loss_event_frequency = MonteCarloSimulation.from_dict(
            values.get("loss_event_frequency")
        )
        risk.ale = MonteCarloSimulation.from_dict(values.get("ale"))
        risk.annual_loss_expectancy = MonteCarloSimulation.from_dict(
            values.get("annual_loss_expectancy")
        )
        return risk

    def __repr__(self):
        return str(self.to_dict())

    def __eq__(self, other):
        if isinstance(other, QuantitativeRisk):
            return self.__hash__() == other.__hash__()
        return False

    def __gt__(self, other):
        if self == other:
            return False
        res = False

    def __hash__(self):
        return hash(
            (
                self.threat_event_frequency,
                self.vuln_score,
                self.loss_event_frequency,
                self.loss_magnitude,
                self.ale,
                self.annual_loss_expectancy,
                self.budget,
                self.currency,
            )
        )

    def __str__(self):
        return str(
            "ALE (p90): "
            + str(round(self.annual_loss_expectancy.p90, 2))
            + f" {self.currency}/år\nALE (p50): "
            + str(round(self.annual_loss_expectancy.probable, 2))
            + f" {self.currency}/år\n"
        )



from typing import Dict, List

from .utils import freeze


class QualitativeScale:
    DEFAULT_NUM_TO_TEXT: Dict[str, int] = {
        5: "Very High",
        4: "High",
        3: "Moderate",
        2: "Low",
        1: "Very Low",
        0: "Very Low",
    }

    DEFAULT_RANGE_TO_TEXT: List[Dict] = [
        {"value": 1, "low": 1, "high": 5},
        {"value": 2, "low": 5, "high": 9},
        {"value": 3, "low": 9, "high": 13},
        {"value": 4, "low": 13, "high": 20},
        {"value": 5, "low": 20, "high": 26},
    ]

    DEFAULT_TEF: List[Dict] = [
        {"value": 1, "low": 0.01, "high": 0.1},
        {"value": 2, "low": 0.1, "high": 1},
        {"value": 3, "low": 1, "high": 10},
        {"value": 4, "low": 10, "high": 100},
        {"value": 5, "low": 100, "high": 1000},
    ]

    DEFAULT_VULNERABILITY: List[Dict] = [
        {"value": 1, "low": 0.01, "high": 0.12},
        {"value": 2, "low": 0.12, "high": 0.25},
        {"value": 3, "low": 0.25, "high": 0.5},
        {"value": 4, "low": 0.5, "high": 0.75},
        {"value": 5, "low": 0.75, "high": 1},
    ]

    DEFAULT_CONSEQUENCE: List[Dict] = [
        {"value": 1, "low": 0.0001, "high": 0.005},
        {"value": 2, "low": 0.005, "high": 0.01},
        {"value": 3, "low": 0.01, "high": 0.02},
        {"value": 4, "low": 0.02, "high": 0.05},
        {"value": 5, "low": 0.05, "high": 1},
    ]

    def __init__(self, scales: dict = None):
        if not scales:
            scales = dict()
        self.likelihood_initiation_or_occurence = scales.get(
            "likelihood_initiation_or_occurence", self.DEFAULT_TEF
        )
        self.likelihood_adverse_impact = scales.get(
            "likelihood_adverse_impact", self.DEFAULT_VULNERABILITY
        )
        self.impact = scales.get("impact", self.DEFAULT_CONSEQUENCE)
        self.num_to_text = scales.get("num_to_text", self.DEFAULT_NUM_TO_TEXT)
        self.risk = scales.get("risk", self.DEFAULT_RANGE_TO_TEXT)

    def to_dict(self):
        return {
            "likelihood_initiation_or_occurence": self.likelihood_initiation_or_occurence,
            "likelihood_adverse_impact": self.likelihood_adverse_impact,
            "impact": self.impact,
            "num_to_text": self.num_to_text,
            "risk": self.risk,
        }

    def __hash__(self):
        return hash(freeze(self.to_dict()))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    @classmethod
    def from_dict(cls, scales: dict):
        tmp = QualitativeScale()
        tmp.likelihood_initiation_or_occurence = scales.get(
            "likelihood_initiation_or_occurence", cls.DEFAULT_TEF
        )
        tmp.likelihood_adverse_impact = scales.get(
            "likelihood_adverse_impact", cls.DEFAULT_VULNERABILITY
        )
        tmp.impact = scales.get("impact", cls.DEFAULT_CONSEQUENCE)
        tmp.num_to_text = scales.get("num_to_text", cls.DEFAULT_NUM_TO_TEXT)
        tmp.risk = scales.get("risk", cls.DEFAULT_RANGE_TO_TEXT)
        return tmp

    def get(self, raw, mapping):
        scale = getattr(self, mapping, None)
        num = None
        if not scale:
            raise AttributeError(f"{mapping} does not exist")
        if raw < scale[0].get("low"):
            num = scale[0].get("value")
        elif raw > scale[-1].get("low"):
            num = scale[-1].get("value")
        else:
            for range in scale:
                if range.get("low") <= raw < range.get("high"):
                    num = range.get("value")
        lookup = num
        if not self.num_to_text.get(num):
            lookup = str(num)
        return {"numeric": num, "text": self.num_to_text.get(lookup)}
